// Copyright 2015, Tobias Hermann and the FunctionalPlus contributors.
// https://github.com/Dobiasd/FunctionalPlus
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)

#pragma once

#include <fplus/benchmark_session.hpp>
#include <fplus/compare.hpp>
#include <fplus/composition.hpp>
#include <fplus/container_common.hpp>
#include <fplus/container_properties.hpp>
#include <fplus/container_traits.hpp>
#include <fplus/extrapolate.hpp>
#include <fplus/filter.hpp>
#include <fplus/generate.hpp>
#include <fplus/interpolate.hpp>
#include <fplus/maps.hpp>
#include <fplus/maybe.hpp>
#include <fplus/numeric.hpp>
#include <fplus/optimize.hpp>
#include <fplus/pairs.hpp>
#include <fplus/queue.hpp>
#include <fplus/raii.hpp>
#include <fplus/read.hpp>
#include <fplus/replace.hpp>
#include <fplus/result.hpp>
#include <fplus/search.hpp>
#include <fplus/sets.hpp>
#include <fplus/shared_ref.hpp>
#include <fplus/show.hpp>
#include <fplus/side_effects.hpp>
#include <fplus/split.hpp>
#include <fplus/stopwatch.hpp>
#include <fplus/string_tools.hpp>
#include <fplus/timed.hpp>
#include <fplus/transform.hpp>
#include <fplus/tree.hpp>
#include <fplus/variant.hpp>

#include <fplus/curry.hpp>
#include <fplus/fwd.hpp>
